/* $Id: bonus.cpp,v 1.14 2005/04/08 13:50:42 pohlt Exp $ */

#include "bonus.hpp"
#include "avatar.hpp"
#include "world.hpp"

/**********************************************************/

#define  ADHESION_MAX_SPEED  5.0
// the default life time of boni in seconds
#define  DEFAULT_LIFE_TIME   120

/**********************************************************/

Bonus::Bonus()
     : m_registered( false ),
       m_updatesPerFrame( 2 ),
       m_numUpdates( m_updatesPerFrame ),
       m_timeToLive( (Sint32)(25 * DEFAULT_LIFE_TIME) )
{
	// default mass for each bonus
	m_mass           = 20.0;
	// default size for the collission rectangle of boni
	m_collRectX      = -7;
	m_collRectY      = -7;
	m_collRectWidth  = 15;
	m_collRectHeight = 13;

	// Each bonus object should be registered first per default,
	// since it is created from class World and therefore exists
	// inside the world. But, unfortunately, we cannot do this
	// here in the constructor of Bonus, because the registration
	// routine calls the virtual function getID(), and at this
	// point of the object generation the bonus is not yet
	// completely built.
}

/**********************************************************/

Bonus::~Bonus() {
}

/**********************************************************/

void Bonus::registerAtManager()
{
	m_worldPointer->getBonusManager()->registerBonusEntrance(*this);
}

/**********************************************************/

void Bonus::unregisterAtManager()
{
	m_worldPointer->getBonusManager()->registerBonusExit(*this);
}

/**********************************************************/

Avatar* Bonus::checkPicking()
{
	Avatar      *avatar   = NULL;
	const Uint8  nPlayers = m_worldPointer->getPlayerList().size();

	// check the avatar of each player
	for( Uint8 p = 0; p < nPlayers; p++ ) {
		const Avatar* const A =
			m_worldPointer->getPlayerList()[p]->getAvatar();
		// only check living avatars and those, who could
		// pick up this bonus at the moment
		if( !A->isLiving() || !canBePickedUpByAvatar(A) )  continue;
		// check the distance to the avatar
		const Sint32 cdx = (Sint32)(A->getPositionX() - getPositionX() + 0.5),
		             cdy = (Sint32)(A->getPositionY() - getPositionY() + 0.5),
		              dx =   cdx > 0
		                   ? (   cdx + A->getCollRectX()
		                       - getCollRectX() - getCollRectWidth())
		                   : ( - cdx + getCollRectX()
		                       - A->getCollRectX() - A->getCollRectWidth() );
		// if the avatar is in long distance with respect to
		// the x-coordinate, we can abort the checking here
		if( dx > 1 ) continue;
		const Sint32  dy =   cdy > 0
		                   ? (   cdy + A->getCollRectY()
		                       - getCollRectY() - getCollRectHeight() )
		                   : ( - cdy + getCollRectY()
		                       - A->getCollRectY() - A->getCollRectHeight() );
		if( dy <= 1 ) {
			if( avatar == NULL || (globalRnd.getUint32() & 1) == 0 ) {
				avatar = (Avatar*)A;
			}
		}
	}

	return avatar;
}

/**********************************************************/

void Bonus::beingPickedUp( const Avatar* const avatar )
{
	m_worldPointer->newObject(BLUEGLOW_25)
		->setPosition( getPositionX(), getPositionY() );
}

/**********************************************************/

void Bonus::update()
{
	// get pointer to the world's map
	Map* const  map = m_worldPointer->getMap();

	const real DT  = 1.0;
	int cx, cy, pcx, pcy;

	// remove the collission rectangle
	removeCollRect();
	// test, whether the object is still laying on ground,
	// or ripped of the ground
	if( isGrounded() && ( ! testGrounded() || getSpeed().abs2() > 0.01) ) {
		setState( GROUNDED, false );
	}
	// if bonus is grounded after this test, set speed to zero
	if( isGrounded() ) {
		m_speedX = m_speedY = 0;
	// if the bonus is not grounded
	} else {
		// move bonus
		m_speedY += GRAVITATION * DT;
		
		real newPosX = m_positionX + (m_speedX * DT),
		      newPosY = m_positionY + (m_speedY * DT);
		
		// is there something to collide?
		if( map->testLineCollRect( ROUND( m_positionX ) + m_collRectX,
		                           ROUND( m_positionY ) + m_collRectY,
		                           ROUND( newPosX     ) + m_collRectX,
		                           ROUND( newPosY     ) + m_collRectY,
		                           m_collRectWidth, m_collRectHeight,
		                           pcx, pcy, cx, cy ) ) {
			// no, then just keep on flying
			m_positionX = newPosX;
			m_positionY = newPosY;
		// otherwise
		} else {
			// move the bonus to the collission point
			m_positionX = pcx - m_collRectX;
			m_positionY = pcy - m_collRectY;

			// do not respect ground interaction in first update step
			if( m_firstUpdate == false ) {
				// squared norm of speed
				const real speed2 =   m_speedX * m_speedX
				                     + m_speedY * m_speedY;
				// if adhesion pins the bonus on ground
				if( speed2 < ADHESION_MAX_SPEED && testGrounded() ) {
					setState( GROUNDED, true );
					m_speedX = m_speedY = 0;
				// otherwise bounce on ground
				} else {
					const real speed    = SQRT_REAL( speed2 ),
					            outAngle = map->calculateBounceAngle(
					                           m_speedX, m_speedY, cx, cy,
					                           Flags::OBSTACLE_EARTH  |
					                           Flags::OBSTACLE_OBJECT   );
					m_speedX = speed * (speed > 5.0 ? 0.5 : 0.9) * COS_REAL( outAngle );
					m_speedY = speed * (speed > 5.0 ? 0.5 : 0.9) * SIN_REAL( outAngle );
				}
			}
		}
	}

	// check destruction of the bonus
	if( m_health <= 0 || --m_timeToLive <= 0 ) {
		// some white particles
		for( int i = 0; i < 100; i++ ) {
			m_worldPointer->getParticles()->addParticle( Particles::ParticleData(
			    getPosition(),
			    globalRnd.getVector( 0.2 + 5*globalRnd.getNormedReal()),// + getSpeed(),
			    Particles::BOUNCE_EARTH | Particles::STAIN | Particles::REDUNDANT,
			    0, 0, 30, 0,
			    globalRnd.getUint32() & 1 ? 0xffffff : 0xdddddd) );			
		}
		// remove bonus from world
		m_worldPointer->deleteObject( this );
		// DO NOT TOUCH ANY BONUS DATA MEMBERS AFTER THIS POINT
		return;
	}

	// check, if the bonus is picked up by an avatar
	Avatar* const avatar = checkPicking();
	// If we found an avatar, who picks up the bonus, pass it
	// to this avatar. This avatar will unhook it from the world's
	// control. The collission rectangle is not set at this moment.
	// Fine, leave the function.
	if( avatar != NULL ) {
		avatar->pickUp( this );
		// IMPORTANT: being picked up by an avatar might delete
		// the bonus object. Therefore do not touch any bonus data
		// members after picking up the bonus
		return;
	}
	
	// set collission rectangle
	setCollRect();
	// standard update of the sprite frame index
	updateSprite();

	m_firstUpdate = false;

	DBG( 5 ) map->setColorAt( ROUND( m_positionX ),
	                          ROUND( m_positionY ), 0xffffff );
}

/**********************************************************/

int Bonus::applyDamage( const Particles::ParticleData& p )
{
	const int damage = CollidableObject::applyDamage( p );

	// generate one white particle
	m_worldPointer->getParticles()->addParticle( Particles::ParticleData(
	    // starting at the position the damaging particle 
	    p.pos,
	    // leaping in a randomly perturbated initial speed vector
	    globalRnd.getVector(1 + 7*globalRnd.getNormedReal()) + getSpeed(),
	    // initial type of the white particle
	    Particles::BOUNCE_EARTH | Particles::STAIN | Particles::REDUNDANT,
	    // owner and modifier
	    0, 0,
	    // time to live
	    globalRnd.getUint32Between( 5, 20 ),
		// this particle does not harm anything
		0,
		// color == WHITE
		0xffffff) );

	return damage;
}

/**********************************************************/

Uint32 Bonus::getSerializeBufferSize() const
{
	return   CollidableObject::getSerializeBufferSize()
	       + Serialize<bool>::sizeOf( m_registered )
	       + Serialize<Sint32>::sizeOf( m_updatesPerFrame )
	       + Serialize<Sint32>::sizeOf( m_numUpdates )
	       + Serialize<Sint32>::sizeOf( m_timeToLive )
	         // adds additional size for debugging (see serialize.hpp),
	         // but only, if the tags are used
	         PLUS_TAG_SIZE( 1 );
}

/**********************************************************/

void Bonus::serialize( Uint8*& bufferPointer ) const
{
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );

	CollidableObject::serialize( bufferPointer );

	Serialize<bool>  ::serialize( m_registered,      bufferPointer );
	Serialize<Sint32>::serialize( m_updatesPerFrame, bufferPointer );
	Serialize<Sint32>::serialize( m_numUpdates,      bufferPointer );
	Serialize<Sint32>::serialize( m_timeToLive,      bufferPointer );

	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, Bonus );
}

/**********************************************************/

void Bonus::deserialize( Uint8*& bufferPointer )
{
	CollidableObject::deserialize( bufferPointer );

	Serialize<bool>  ::deserialize( bufferPointer, m_registered );
	Serialize<Sint32>::deserialize( bufferPointer, m_updatesPerFrame );
	Serialize<Sint32>::deserialize( bufferPointer, m_numUpdates );
	Serialize<Sint32>::deserialize( bufferPointer, m_timeToLive );

	// expands to tag deserialization
	DESERIALIZE_OBJECT_TAG( bufferPointer );
}

/**********************************************************/

void Bonus::updateSprite()
{
	if( --m_numUpdates <= 0 ) {
		if( ++m_Frame >= m_SequenceLength )  m_Frame = 0;
		m_numUpdates = m_updatesPerFrame;
	}
}

/**********************************************************/
