/* $Id: client.cpp,v 1.165 2005/05/05 11:52:20 chfreund Exp $ */

#include "client.hpp"

#include <cmath>
#include <set>
#include <sstream>

#include "global.hpp"
#include "event.hpp"
#include "video.hpp"
#include "audio.hpp"
#include "avatar.hpp"
#include "particles.hpp"
#include "signalhandler.hpp"
#include "spritesequence.cpp"
#include "spriteset.cpp"
#include "wopsprites.cpp"
#include "flag.hpp"
#include "rope.hpp"

/**********************************************************/

Client::Client()
	: m_comm( this ), m_world( NULL ), m_status( DEFAULT ), m_background( NULL ),
	  m_player( NULL ), m_gui( NULL ),
	  m_viewX( 0 ), m_viewY( 0 ), m_viewVelX( 0 ), m_viewVelY( 0 ),
	  m_doScreenshot( false ) { 
	ASSERT( !SDL_InitSubSystem( SDL_INIT_TIMER ),
	        "Client::Client: Could not initialize timer\n" );

	// init loader video object
	Loader::setVideo( &m_video );

	// create GUI for the client
	m_gui = new WopGUI( this, m_video.getScreen() );
	DBG( 1 ) ASSERT( m_gui != NULL, "Client::Client: couldn't initialize gui\n" );
}

/**********************************************************/

Client::~Client() {
	LOG( 1 ) INFO( "Client::~Client: starting destructor\n" );

	if( m_gui ) {
		delete m_gui;
		m_gui = 0;
	}
	if( m_world ) {
		delete m_world;
		m_world = 0;
	}
	if( m_background ) {
		delete m_background;
		m_background = 0;
	}

	LOG( 1 ) INFO( "Client::~Client: done\n" );
}

/**********************************************************/

void Client::run() {
	init();
	eventLoop();
	LOG( 2 ) INFO( "Client::run: exiting\n" );
}

void Client::eventLoop() {
	SignalHandler* signalHandler = SignalHandler::getInstance();
	while ( ! signalHandler->hasQuit() && m_status != SHUTDOWN ) {
		SDL_Event event;
		SDL_WaitEvent( &event );
		// first try if the GUI can consume this event otherwise
		// let's try it ourselves
		if ( ! m_gui->handleEvent( &event ) ) {
			handleEvent( &event );
		}
		m_audio->update();
	}
	
	LOG( 2 ) INFO( "Client::eventLoop: starting shutdown\n" );
	m_status = SHUTDOWN;
	SDL_WaitThread( m_displayThread, 0 );
	LOG( 3 ) INFO( "Client::eventLoop: joined display thread\n" );
	SDL_DestroyMutex( m_displayLoopMutex );
	SDL_DestroyCond( m_displayLoopCond );
}

void Client::exit() {
	if ( m_comm.isConnected() )
		disconnect();
	else
		m_status = SHUTDOWN;
}

/**********************************************************/

void Client::init()
{
	m_showWorld = false;
	
	SDL_Surface* wmIcon =
		Loader::getInstance()->getImage( "images/gui/wop32.png", true );
	SDL_WM_SetIcon( wmIcon, 0 );
	
	// get WoP configuration
	WopSettings* const settings = WopSettings::getInstance();
	// Initialize audio
	m_audio = Audio::getInstance();
	m_audio->loadTrack( 0 ); // menu track
	m_audio->playTrack();

	// set cursor
	m_video.setMouseCursor( "   +       +   " //  1
	                        "   +       +   " //  2
	                        "    +     +    " //  3
	                        "++  +     +  ++" //  4
	                        "  ++ +   + ++  " //  5
	                        "    ++   ++    " //  6
	                        "               " //  7
	                        "       #       " //  8
	                        "               " //  9
	                        "    ++   ++    " // 10
	                        "  ++ +   + ++  " // 11
	                        "++  +     +  ++" // 12
	                        "    +     +    " // 13
	                        "   +       +   " // 14
	                        "   +       +   ", 15, 15, 7, 7 );
	m_video.showMouseCursor( true );

	// reset logging object and initialize it
	m_ProgressLog.reset();
	m_ProgressLog.create( 30, 1 );
	
	m_displayLoopMutex = SDL_CreateMutex();
	m_displayLoopCond  = SDL_CreateCond();
	m_displayThread = SDL_CreateThread( displayLoopWrapper, this );
	
	m_gui->showSplashScreen();
	
	// load all sprites (if this does not succeed,
	// the program is terminated)
	m_gui->showInitializationWindow();
	m_gui->setInitializationValue( static_cast<Uint32>( 100.0 * 0 ) );
	m_video.loadSprites( &m_ProgressLog );
	m_gui->hideInitializationWindow();

	/*
	// create player object
	m_player = NEW Player();
	ASSERT( m_player, "Client::init: Could not create player\n" );
	// some attributes for the player
	m_player->setName( settings->getName() );
	m_player->setPlayerColor( settings->getPlayerColor() );
	m_player->setTeamID( settings->getTeamID() );
	*/
	
	// initialize keys
	if ( !strcmp( settings->getKeyboard(), "de" ) ) {
		m_ropeOnKey = SDLK_y;
	}
	else {
		m_ropeOnKey = SDLK_z;
	}
	
	// hide splash screen
	m_gui->hideSplashScreen();
	
	m_frame = 0;
	
	// if a server name is given, try to connect to this one
	if ( Setting* s = settings->getSetting( "server" ) ) {
		ServerEntry server( s->getString( 0 ) );
		// connect to the server
		connectToServer( server );
	}
	else {
		// std::set<ServerEntry> servers;
		// servers = m_comm.lookForServer();
		// m_gui->showServerListWindow( servers );
		m_gui->showMainMenu();
	}
}

/**********************************************************/

bool Client::connectToServer( ServerEntry& server ) {
	// create a new player
	m_player = NEW Player();
	ASSERT( m_player, "Client::connectToServer: Could not create player\n" );
	// some attributes for the player
	WopSettings* settings = WopSettings::getInstance();
	m_player->setName( settings->getName() );
	m_player->setPlayerColor( settings->getPlayerColor() );
	m_player->setTeamID( settings->getTeamID() );
	
	m_status = SYNC;
	
	// try to establish a connection to the server
	if ( !m_comm.setServer( server ) )  return false;
	m_audio->fadeOutTrack( 10000 );
	m_audio->setJukeboxMode( true );
	// this is the place the world is created and m_world is assigned
	m_comm.connect( m_player, m_world );
	// We need a pointer to the player object in the world so we
	// delete the current player object and pick the last player
	// in the world received... this can probably be done better
	delete m_player;
	m_player = m_world->getPlayer( m_world->getNumberPlayers() - 1 );
	// mark this player as local player
	m_player->setLocalPlayer();
	
	for ( Uint8 i = 0; i < m_world->getNumberPlayers(); i++ ) {
		m_world->getPlayer( i )->getAvatar()->setMessageSink( this );
	}
	m_world->setMessageSink( this );
	
	// connect the sprite class in the video class with the world
	// using the interface base class of WopSprites
	LOG( 2 ) INFO( "Client::connect: attaching sprite interface to world\n" );
	DBG( 1 ) ASSERT( m_video.getSpriteInterface(),
	                 "Client::connect: m_video.getSpriteInterface() == NULL\n" );
	m_world->setSpriteInterface( m_video.getSpriteInterface() );
	
	// create a surface from the pixel values in the map
	m_world->optimizeForClient( m_video.getScreen()->format, SDL_SWSURFACE );
	ClientMap* clientMap = dynamic_cast<ClientMap*>( m_world->getMap() );
	ASSERT( clientMap, "Client::connect: did not get ClientMap\n" );
	m_mapSurface = clientMap->getPixels();
	ASSERT( m_mapSurface, "Client::connect: Could not create map surface (%s)\n",
	        SDL_GetError() );
	SDL_SetColorKey( m_mapSurface, SDL_SRCCOLORKEY, 0 );

	m_mapRect.x = m_mapRect.y = 0;
	m_mapRect.w = m_video.getScreen()->w;
	m_mapRect.h = m_video.getScreen()->h;

	m_background = new WopBackground( m_world->getMap()->getSizeX(),
	                                  m_world->getMap()->getSizeY(),
	                                  "images/mapstuff/background/dark_earth.jpg",
	                                  m_video.getScreen()->format );
	
	m_gui->showGameWidgets();
	
	m_status = DEFAULT;
	m_showWorld = true;
	// m_sendTimer = SDL_AddTimer( 20, sendTimerCallback, this );
	m_frameTimer = SDL_AddTimer( 40, frameTimerCallback, this );
	
	return true;
}

void Client::disconnect() {
	LOG( 2 ) INFO( "Client::disconnect: disconnecting...\n" );
	LOG( 3 ) INFO( "Client::disconnect: removing frame timer\n" );
	// m_status = SHUTDOWN;
	SDL_RemoveTimer( m_frameTimer );
	// LOG( 3 ) INFO( "Client::disconnect: waiting for frame thread to finish\n" );
	// SDL_Delay( 500 );
	LOG( 3 ) INFO( "Client::disconnect: disconnecting from server\n" );
	m_comm.disconnect();

	m_audio->setJukeboxMode( false );
	m_audio->loadTrack( 0 );
	m_audio->fadeInTrack( 3000 );

	// reset all connection dependent objects
	m_gui->hideGameWidgets();
	SDL_mutexP( m_displayLoopMutex );
	m_showWorld = false;
	SDL_mutexV( m_displayLoopMutex );
	delete m_background;
	m_background = 0;
	delete m_world;
	m_world = 0;
	
	std::set<ServerEntry> servers;
	servers = m_comm.lookForServer();
	m_gui->showServerListWindow( servers );
}

bool Client::handleEvent( SDL_Event* event ) {
	// the events the client is interested in are the key events for
	// controlling the player's avatar, ENTER for writing a chat message
	// and ESCAPE for leaving the game
	// and much more now (Tom)
	if ( event->type == SDL_KEYDOWN ) {
		switch( event->key.keysym.sym ) {
		case SDLK_ESCAPE:
			exit();
			break;
		case SDLK_RETURN:
			if ( m_comm.isConnected() ) {
				SDL_EnableUNICODE( true );
				m_gui->showChatWindow();
			}
			break;
		case SDLK_LEFT:
			m_myEvent.addAction( Event::LEFT );
			break;
		case SDLK_RIGHT:
			m_myEvent.addAction( Event::RIGHT );
			break;
		case SDLK_DOWN:
			m_myEvent.addAction( Event::DOWN );
			break;
		case SDLK_UP:
			m_myEvent.addAction( Event::UP );
			break;
		case SDLK_LCTRL:
			m_myEvent.addAction( Event::SHOOT );
			break;
		case SDLK_LSHIFT:
			m_myEvent.addAction( Event::JUMP );
			break;
		case SDLK_LALT:
			m_myEvent.addAction( Event::DIG );
			break;
		case SDLK_f:
			m_myEvent.addAction( Event::JET );
			break;
		// case SDLK_z:
		// 	m_myEvent.addAction( Event::ROPE_ON );
		// 	break;
		case SDLK_x:
			m_myEvent.addAction( Event::ROPE_OFF );
			break;
		case SDLK_a:
			m_myEvent.addAction( Event::ROPE_IN );
			break;
		case SDLK_s:
			m_myEvent.addAction( Event::ROPE_OUT );
			break;
		case SDLK_d:
			m_myEvent.addAction( Event::ROPE_REL );
			break;
		case SDLK_o:
			m_myEvent.addAction( Event::SCREENSHOT );
			break;
		case SDLK_F12:
			m_myEvent.addAction( Event::DEBUG );
			break;
		case SDLK_F5:
			m_audio->decreaseTrackVolume();
			break;
		case SDLK_F6:
			m_audio->increaseTrackVolume();
			break;
		case SDLK_F7:
			if ( m_comm.isConnected() ) m_audio->toggleJukeboxMode();
			break;
		case SDLK_1:
			m_myEvent.setWeapon( 1 );
			break;
		case SDLK_2:
			m_myEvent.setWeapon( 2 );
			break;
		case SDLK_3:
			m_myEvent.setWeapon( 3 );
			break;
		case SDLK_4:
			m_myEvent.setWeapon( 4 );
			break;
		case SDLK_5:
			m_myEvent.setWeapon( 5 );
			break;
		case SDLK_6:
			m_myEvent.setWeapon( 6 );
			break;
		case SDLK_7:
			m_myEvent.setWeapon( 7 );
			break;
		case SDLK_8:
			m_myEvent.setWeapon( 8 );
			break;
		case SDLK_SPACE:
			m_myEvent.addAction( Event::SPAWN );
			break;
		case SDLK_v:
			m_myEvent.setWeapon( ( ( m_player->getActiveWeaponIndex() + 1 ) % (Player::m_MAX_WEAPONS-Player::m_IGNORE_N_LAST_WEAPONS) ) + 1 );
			break;
		case SDLK_c:
			m_myEvent.setWeapon( ( ( m_player->getActiveWeaponIndex() + Player::m_MAX_WEAPONS-Player::m_IGNORE_N_LAST_WEAPONS-1 ) % (Player::m_MAX_WEAPONS-Player::m_IGNORE_N_LAST_WEAPONS) ) + 1 );
			break;
		default:
			// nothing to do here
			break;
		}
		if ( event->key.keysym.sym == m_ropeOnKey ) {
			m_myEvent.addAction( Event::ROPE_ON );
		}
	}
	else if ( event->type == SDL_KEYUP ) {
		switch( event->key.keysym.sym ) {
		case SDLK_LEFT:
			m_myEvent.removeAction( Event::LEFT );
			break;
		case SDLK_RIGHT:
			m_myEvent.removeAction( Event::RIGHT );
			break;
		case SDLK_DOWN:
			m_myEvent.removeAction( Event::DOWN );
			break;
		case SDLK_UP:
			m_myEvent.removeAction( Event::UP );
			break;
		case SDLK_LCTRL:
			m_myEvent.removeAction( Event::SHOOT );
			break;
		case SDLK_LSHIFT:
			m_myEvent.removeAction( Event::JUMP );
			break;
		case SDLK_LALT:
			m_myEvent.removeAction( Event::DIG );
			break;
		case SDLK_f:
			m_myEvent.removeAction( Event::JET );
			break;
		// case SDLK_z:
		// 	m_myEvent.removeAction( Event::ROPE_ON );
		// 	break;
		case SDLK_x:
			m_myEvent.removeAction( Event::ROPE_OFF );
			break;
		case SDLK_a:
			m_myEvent.removeAction( Event::ROPE_IN );
			break;
		case SDLK_s:
			m_myEvent.removeAction( Event::ROPE_OUT );
			break;
		case SDLK_d:
			m_myEvent.removeAction( Event::ROPE_REL );
			break;
		case SDLK_o:
			m_myEvent.removeAction( Event::SCREENSHOT );
			break;
		case SDLK_F12:
			m_myEvent.removeAction( Event::DEBUG );
			break;
		case SDLK_1:
		case SDLK_2:
		case SDLK_3:
		case SDLK_4:
		case SDLK_5:
		case SDLK_6:
		case SDLK_7:
		case SDLK_8:
		case SDLK_c:
		case SDLK_v:
			m_myEvent.removeWeapon();
			break;
		case SDLK_SPACE:
			m_myEvent.removeAction( Event::SPAWN );
			break;
		default:
			// nothing to do here
			break;
		}
		if ( event->key.keysym.sym ==  m_ropeOnKey ) {
			m_myEvent.removeAction( Event::ROPE_ON );
		}
	}
	
	return false;
}

/**********************************************************/

bool Client::receiveAndProcessFrame( Uint8 frame ) {
	// wait for the next echo message
	EchoData* echoData = m_comm.getEcho( frame );

	// Quick return if no echo could be received
	if ( !echoData )
		return false;

	// Check if the random counter from the server matches our counter
	if ( echoData->randomCounter != globalRnd.getCounter() ) {
		INFO( "ERROR: Client is no longer synchronous with the server" );
		disconnect();
		return false;
	}
	
	// send own event to the server
	m_comm.sendEvent( m_myEvent );
	// INFO( "Sent event for frame %d: %x\n", m_frame, m_myEvent.get() );

	// process echo message
	m_world->applyEvents( echoData );
	m_world->doTimestep();
	
	for ( Uint8 c = 0; c < m_world->getNumberPlayers(); c++ ) {
		// SCREENSHOT triggers screenshot capture for debugging
		Event event = echoData->event[c];
		if( event.testAnyActions( Event::SCREENSHOT )) {
			m_doScreenshot = true;
		}

		// DEBUG paints the flag field as the background
		if( event.testAnyActions( Event::DEBUG )) {
			m_world->getMap()->displayFlags();
			m_world->getMap()->saveFlags( "DEBUG event" );
		}
	}
	// cleanup and return
	delete echoData;
	return true;
}

/**********************************************************/

void Client::processEvents() {
	// Check for synchronisation needs
	m_comm.handleTCPMessages();
	
	// process echos until no more are available
	while ( receiveAndProcessFrame( m_frame ) ) {
		m_frame++;
	}
	
	// tell the display thread to update the graphics
	SDL_CondSignal( m_displayLoopCond );
}

/**********************************************************/

void Client::addPlayer( Player* player, Uint8 frame )
{
	// Very bad if we are in a later frame than the on where the new player should enter
	ASSERT ( frame >= m_frame,
	         "Client::addPlayer: Client frame %d is larger than player frame %d\n",
	         m_frame, frame );
	
	// Receive all frames up to the one where the new player enters the world
	while ( m_frame < frame ) {
		if ( receiveAndProcessFrame( m_frame ) ) {
			m_frame++;
		}
	}
	
	// Add the new player to the world and display a info message
	m_world->addPlayer( player );
	String message = "Player ";
	message += player->getName() + " joined the game";
	SDL_Color white;
	white.r = white.g = white.b = 255;
	showMessage( message, white );
	
	player->getAvatar()->setMessageSink( this );
}

void Client::removePlayer( Uint8 playerIndex, Uint8 frame )
{
	// It should not happen that we have skipped over the frame where the player
	// leaves the world... well, let's display a warning
	CHECK ( frame <= m_frame + 1,
	        "Client::removePlayer: Client frame %d is larger than player frame %d\n",
	         m_frame + 1, frame );
	
	// Receive all events up the needed frame
	while ( m_frame <= frame ) {
		if ( receiveAndProcessFrame( m_frame ) ) {
			m_frame++;
		}
	}
	
	LOG( 2 ) INFO( "Client::removePlayer: removing player %d in frame %d\n", playerIndex, frame );
	
	// Remove the player from the world and display a message
	Player* player = m_world->getPlayer( playerIndex );
	String message = "Player ";
	message += player->getName() + " has left the game";
	m_world->removePlayer( playerIndex, frame );
	SDL_Color white;
	white.r = white.g = white.b = 255;
	showMessage( message, white );
	delete player;
}

/**********************************************************/

void Client::sendChatMessage( String& message ) {
	String displayedMessage = m_player->getName() + ": " + message;
	SDL_Color color;
	color.r = TPLCOL_RED( m_player->getPlayerColor() );
	color.g = TPLCOL_GREEN( m_player->getPlayerColor() );
	color.b = TPLCOL_BLUE( m_player->getPlayerColor() );
	m_comm.sendChatMessage( displayedMessage, color );
	m_gui->showMessage( displayedMessage, color );
}

void Client::addMessage( std::string s ) {
	SDL_Color white;
	white.r = white.g = white.b = 255;
	String message( s.c_str() );
	m_gui->showMessage( message, white );
}

void Client::showMessage( String& message, const SDL_Color& color ) {
	m_gui->showMessage( message, color );
}

/**********************************************************/

static inline bool checkAccel( const real dist, const real vel,
                               const real accel, const real decel ) {
	const int N = ROUND( (FABS_REAL( vel )+accel)/decel + 0.9999 );
	return 0.5*N*(N+1)*decel <= FABS_REAL( dist );
}

/**********************************************************/

void Client::updateGraphics()
{
	if ( m_showWorld ) {
		const real viewAccel = 1.5, viewDecel = 0.5;
		m_mapRect.w = m_video.getScreen()->w;
		m_mapRect.h = m_video.getScreen()->h;
		
		/////////////////////////////////////
		// arrange inert moving of viewport
		if( (m_viewVelX >= 0 || checkAccel( m_viewX, m_viewVelX, viewAccel, viewDecel )) &&
		    (m_viewVelX <= 0 || checkAccel( m_mapSurface->w - m_viewX - m_mapRect.w, m_viewVelX, viewAccel, viewDecel )) &&
		    checkAccel( m_viewX + m_mapRect.w/2 - m_player->getAvatar()->getIntPositionX(), m_viewVelX, viewAccel, viewDecel ) ) {
			if     ( m_viewX + m_mapRect.w/2 - m_player->getAvatar()->getIntPositionX() > 0 ) m_viewVelX -= viewAccel;
			else if( m_viewX + m_mapRect.w/2 - m_player->getAvatar()->getIntPositionX() < 0 ) m_viewVelX += viewAccel;
			//INFO( "ACCEL X\n" );
		}
		else {
			if     ( m_viewVelX > 0) m_viewVelX -= viewDecel;
			else if( m_viewVelX < 0) m_viewVelX += viewDecel;
			//INFO( "DECEL X\n" );
		}

		if( (m_viewVelY >= 0 || checkAccel( m_viewY, m_viewVelY, viewAccel, viewDecel )) &&
		    (m_viewVelY <= 0 || checkAccel( m_mapSurface->h - m_viewY - m_mapRect.h, m_viewVelY, viewAccel, viewDecel )) &&
		    checkAccel( m_viewY + m_mapRect.h/2 - m_player->getAvatar()->getIntPositionY(), m_viewVelY, viewAccel, viewDecel ) ) {
			if     ( m_viewY + m_mapRect.h/2 - m_player->getAvatar()->getIntPositionY() > 0 ) m_viewVelY -= viewAccel;
			else if( m_viewY + m_mapRect.h/2 - m_player->getAvatar()->getIntPositionY() < 0 ) m_viewVelY += viewAccel;
			//INFO( "ACCEL Y\n" );
		}
		else {
			if     ( m_viewVelY > 0) m_viewVelY -= viewDecel;
			else if( m_viewVelY < 0) m_viewVelY += viewDecel;
			//INFO( "DECEL Y\n" );
		}

		m_viewX += m_viewVelX;
		m_viewY += m_viewVelY;

		if( m_viewX < 0 ) {
			//INFO( "m_viewX < 0\n" );
			m_viewX = 0; m_viewVelX = 0;
		}
		if( m_viewX > m_mapSurface->w - m_mapRect.w + 80 ) {
			//INFO( "m_viewX > X\n" );
			m_viewX = m_mapSurface->w - m_mapRect.w + 80; m_viewVelX = 0;
		}
		if( m_viewY < 0 ) {
			//INFO( "m_viewY < 0\n" );
			m_viewY = 0; m_viewVelY = 0;
		}
		if( m_viewY > m_mapSurface->h - m_mapRect.h + 50 ) {
			//INFO( "m_viewY > Y\n" );
			m_viewY = m_mapSurface->h - m_mapRect.h + 50; m_viewVelY = 0;
		}
		m_mapRect.x = ROUND( m_viewX );
		m_mapRect.y = ROUND( m_viewY );
		if ( m_mapRect.x > m_mapSurface->w - m_video.getScreen()->w ) {
			m_mapRect.w = m_mapSurface->w - m_mapRect.x;
			SDL_Rect clearRect;
			clearRect.x = m_mapRect.w;
			clearRect.y = 0;
			clearRect.w = m_video.getScreen()->w - m_mapRect.w;
			clearRect.h = m_video.getScreen()->h;
			SDL_FillRect( m_video.getScreen(), &clearRect, 0 );
		}
		if ( m_mapRect.y > m_mapSurface->h - m_video.getScreen()->h ) {
			m_mapRect.h = m_mapSurface->h - m_mapRect.y;
			SDL_Rect clearRect;
			clearRect.x = 0;
			clearRect.y = m_mapRect.h;
			clearRect.w = m_video.getScreen()->w;
			clearRect.h = m_video.getScreen()->h - m_mapRect.h;
			SDL_FillRect( m_video.getScreen(), &clearRect, 0 );
		}
		// end: inert moving of viewport
		/////////////////////////////////////

		// set viewport position for stereo audio
		m_audio->setListeningPosition( m_player->getAvatar()->getIntPositionX(),
		                               m_player->getAvatar()->getIntPositionY() );

		// draw the background image
		SDL_BlitSurface( m_background->getGraphic(), &m_mapRect, m_video.getScreen(), 0 );

		// draw the map
		SDL_BlitSurface( m_mapSurface, &m_mapRect, m_video.getScreen(), NULL );

		// draw the world
		m_world->draw( m_video.getScreen(), m_video.getSprites(), m_mapRect );
	
		// draw aiming target
		if( m_player->getAvatar()->isLiving() ) {
			const int x = ROUND( m_player->getAvatar()->getPositionX() +
											25 * m_player->getAvatar()->getAimingDX() ) - m_mapRect.x;
			const int y = ROUND( m_player->getAvatar()->getPositionY() +
											25 * m_player->getAvatar()->getAimingDY() ) + m_player->getAvatar()->getWeaponPivotDY() - m_mapRect.y;
			if( x >= 1 && y >= 1 && x < m_mapRect.w-1 && y < m_mapRect.h-1 ) {
				Uint32 color = SDL_MapRGB( m_video.getScreen()->format, 255, 255, 255 );
				Graphics::setPixel( m_video.getScreen(), x  , y-1, color );
				Graphics::setPixel( m_video.getScreen(), x-1, y  , color );
				Graphics::setPixel( m_video.getScreen(), x  , y  , color );
				Graphics::setPixel( m_video.getScreen(), x+1, y  , color );
				Graphics::setPixel( m_video.getScreen(), x  , y+1, color );
			}
		}
		
		m_gui->lock();
		
		// set weapon indicators
		for ( int i = 0; i < Player::m_MAX_WEAPONS-Player::m_IGNORE_N_LAST_WEAPONS; i++ ) {
			Weapon* weapon = m_player->getWeapon( i );
			if ( weapon->isCharging() ) {
				SDL_Color chargeColor;
				chargeColor.r = 217; chargeColor.g = 167; chargeColor.b = 167;
				if ( weapon != m_player->getActiveWeapon() ) {
					chargeColor.r /= 2; chargeColor.g /= 2; chargeColor.b /= 2;
				}
				else {
					m_gui->setActiveWeaponColor( chargeColor );
					m_gui->setActiveWeaponValue( weapon->getChargingPercentage() );
				}
				m_gui->setWeaponColor( i, chargeColor );
				m_gui->setWeaponValue( i, weapon->getChargingPercentage() );
			}
			else {
				SDL_Color weaponColor;
				weaponColor.r = 167; weaponColor.g = 167; weaponColor.b = 217;
				if ( weapon != m_player->getActiveWeapon() ) {
					weaponColor.r /= 2; weaponColor.g /= 2; weaponColor.b /= 2;
				}
				else {
					m_gui->setActiveWeaponColor( weaponColor );
					m_gui->setActiveWeaponValue( ( weapon->getCurrentAmmo() * 100 ) / weapon->getMaximumAmmo() );
				}
				m_gui->setWeaponColor( i, weaponColor );
				m_gui->setWeaponValue( i, ( weapon->getCurrentAmmo() * 100 ) / weapon->getMaximumAmmo() );
			}
		}
		
		// update health indicator
		m_gui->setHealthValue( m_player->getAvatar()->getHealth() / 10000 );
		// update rope indicator
		Weapon* ropeWeapon = m_player->getWeapon( Player::m_MAX_WEAPONS-Player::m_IGNORE_N_LAST_WEAPONS );
		if ( ropeWeapon->isCharging() )
			m_gui->setRopeValue( ropeWeapon->getChargingPercentage() );
		else
			m_gui->setRopeValue( ( ropeWeapon->getCurrentAmmo() * 100 ) / ropeWeapon->getMaximumAmmo() );
		// update fuel indicator
		m_gui->setFuelValue( max( 0, m_player->getAvatar()->getFuel()*100 )/Avatar::MAX_FUEL );
		
		// update scores
		Uint8* keyState = SDL_GetKeyState( NULL );
		if ( keyState[SDLK_TAB] ) {
			m_gui->updateScores( m_world->getAllPlayers() );
			m_gui->showScoreTable();
		}
		else {
			m_gui->hideScoreTable();
		}
		
		m_gui->unlock();
	}
	else {
		// just clear the screen
		Uint32 colorBlack = SDL_MapRGB( m_video.getScreen()->format, 0, 0, 0 );
		SDL_FillRect( m_video.getScreen(), 0, colorBlack );
	}
	
	// update initialization
	DBG( 2 ) {
		ASSERT( m_ProgressLog.getNumLevels() > 0,
	            "Client::updateGraphics: progress logging object "
	            "not initialized\n" );
	}
	m_gui->lock();
	m_gui->setInitializationValue( static_cast<Uint32>( 100.0 * m_ProgressLog.getProgress() ) );
	std::string initString( m_ProgressLog.getTitle().getString() );
	m_gui->setInitializationTitle( initString );
	m_gui->unlock();
	// draw the GUI stuff
	m_gui->draw();
	
	SDL_Flip( m_video.getScreen() );
}

/**********************************************************/

void Client::displayLoop() {
	Uint32 frame_counter = 0;
	int tick_reference = SDL_GetTicks();
	int tick_counter = 0;
	while( m_status != SHUTDOWN ) {
		SDL_mutexP( m_displayLoopMutex );

		LOG( 5 ) {
			CHECK( SDL_CondWaitTimeout( m_displayLoopCond, m_displayLoopMutex, 40 ) == 0,
			       "Client::displayLoop: timeout when waiting for display condition\n" );
		}
		else {
			SDL_CondWaitTimeout( m_displayLoopCond, m_displayLoopMutex, 40 );
		}

		updateGraphics();

		// is there a pending screenshot?
		if( m_doScreenshot ) {
			static int screenshotNumber = 0;
			String filename;
			filename.format( "wop_%02i_%02i.bmp", m_player->getPlayerID(),
			                 screenshotNumber++ );
			m_video.saveScreenshot( filename );

			m_doScreenshot = false;
		}

		frame_counter++;
		tick_counter = SDL_GetTicks();
		if ( tick_counter - tick_reference >= 5000 ) {
			LOG( 3 ) INFO( "Client::displayLoop: drawing with about %4.1f fps\n",
			               frame_counter/5.0 );
			frame_counter = 0;
			tick_reference = tick_counter;
		}
		SDL_mutexV( m_displayLoopMutex );
	}
	LOG( 2 ) INFO( "Client::displayLoop: exiting display loop\n" );
}

/**********************************************************/

int displayLoopWrapper( void* data ) {
	Client* clientPtr = static_cast<Client*>( data );
	clientPtr->displayLoop();
	return 0;
}

/**********************************************************/

bool Client::waitForAnyKeyPressed( const int tenthsOfSecond ) {
	for( int i = 0; i < tenthsOfSecond; i++ ) {
		if( m_input.anyKeyPressed() ) return true;
		SDL_Delay( 100 );
	}
	return false;
}		

/**********************************************************/

Uint32 frameTimerCallback( Uint32 interval, void* data ) {
	// INFO( "Start callback (interval: %d)\n", interval );
	Uint32 ticksBefore = SDL_GetTicks();
	Client* client = static_cast<Client*>( data );
	if( client->getStatus() != SHUTDOWN ) client->processEvents();
	Uint32 ticksAfter = SDL_GetTicks();
	Uint32 duration = ticksAfter - ticksBefore;
	Uint32 newInterval = ( duration >= 40 ) ? 1 : 40 - duration;
	// INFO( "End callback (duration: %d)\n", duration );
	return newInterval;
}

