/* $Id: particles.cpp,v 1.41 2005/04/28 12:27:52 pohlt Exp $ */

#include <cmath>

#include "particles.hpp"

#include "constants.hpp"
#include "world.hpp"
#include "global.hpp"
#include "serialize.hpp"


Particles::Particles( World& world ) :
		m_nParticles( 0 ),
		m_world( world ) {
	int c;
	
#define GEN_COLOR(r, g, b) (((((r) << 8) | (g)) << 8) | (b))
	for( c = 0; c < m_N_SPARK_COLORS; c++ ) {
		if     ( c <= 16 ) m_sparkColor[m_N_SPARK_COLORS-1-c] = GEN_COLOR( 255, 255, 255-15*c );
		else if( c <= 33 ) m_sparkColor[m_N_SPARK_COLORS-1-c] = GEN_COLOR( 255, 255-15*(c-17), 0 );
		else               m_sparkColor[m_N_SPARK_COLORS-1-c] = GEN_COLOR( 255-15*(c-34), 0, 0 );
	}
#undef GEN_COLOR
}


void Particles::doTimestep() {
	const real DT = 1;
	int p, pcx = 0, pcy = 0, cx, cy;
	Vector oldPos;
	Map* const map = m_world.getMap();

	for( p = 0; p < m_nParticles; p++ ) {
		// 'part' is just a reference to the current particle
		ParticleData& part = m_particleData[p];

		oldPos = part.pos;
		part.vel.y += GRAVITATION*DT;
		part.pos += part.vel*DT;

		if( (part.type & BOUNCE_OBSTACLE_ANY) &&
		    ! map->testLinePassable( ROUND( oldPos.x   ),
		                             ROUND( oldPos.y   ),
		                             ROUND( part.pos.x ),
		                             ROUND( part.pos.y ),
		                             pcx, pcy, cx, cy,
		                             part.type & BOUNCE_OBSTACLE_ANY )) {

			// if particle stains earth
			if( part.type & STAIN ) {
				map->setColorAt( cx, cy, part.color );
				
				// remove particle
				removeParticle( &p );
				continue;
			}

			// if the particle hit something that makes it explode
			if( ((part.type & EXPL_OBSTACLE_ANY) >> 3) &
			    map->getFlags()->getValueAt( cx, cy ) ) {

				// if particle damages
				if( (part.type & DAMAGE) &&
				    (map->getFlags()->getValueAt( cx, cy ) & Flags::OBSTACLE_OBJECT) ) {
					part.pos.set( cx, cy );
					m_world.damageObject( part );
				}
				else {
					// make hole
					map->makeHole( cx, cy, 3, Flags::INDESTRUCTIBLE );

					if( part.type & EXPL_GEN_SPARKS ) {
						// create the explosion effect
						m_world.newObject(EXPLOSION_15)->setPosition( cx, cy );

						// create sparks
						for( int i = 0; i < 5; i++ ) {
							addParticle( ParticleData( Vector( pcx, pcy ),
							     globalRnd.getVector( 1 + 7*globalRnd.getNormedReal() ),
							     BOUNCE_EARTH | BOUNCE_OBJECT | REDUNDANT | SPARK, 0, 0,
							     m_N_SPARK_COLORS/2 + (globalRnd.getUint32() % (m_N_SPARK_COLORS/2+1)),
							     0, 0xffffff));
						}
					}
				}

				// remove particle
				removeParticle( &p );
				continue;
			}
			// do not explode, just bounce off
			else {
				real speed = part.vel.abs();

				if( part.type & SPECIAL_BOUNCE ) {
					speed *= 0.01 * static_cast<real>( part.modifier );
				}
				else {
					speed *= (speed < 1.5 ? 1.0 : 0.6);
				}

				part.vel.setAngle( map->calculateBounceAngle( part.vel.x, part.vel.y,
				             cx, cy, part.type & BOUNCE_OBSTACLE_ANY ), speed );
				part.pos.set( pcx, pcy );
			}
		}

		// if out of map
		if( part.pos.x < 0 || part.pos.x >= map->getSizeX() ||
		    part.pos.y < 0 || part.pos.y >= map->getSizeY() ) {
			CHECK( false, "Particles::doTimestep: particle out of map\n" );
			removeParticle( &p );
			continue;
		}

		// if TTL relevant (>= 0) and TTL exceeded
		else if( part.ttl >= 0 && --part.ttl <= 0 ) {
			if( (part.type & EXPL_TTL) && (part.type & EXPL_GEN_SPARKS) ) {
				// create the explosion effect
				m_world.newObject(EXPLOSION_15)->setPosition( part.pos.x + 0.5,
				                                              part.pos.y + 0.5 );
				for( int i = 0; i < 5; i++ ) {
					addParticle( ParticleData( Vector( pcx, pcy ),
					     globalRnd.getVector( 1 + 7*globalRnd.getNormedReal() ),
					     BOUNCE_EARTH | BOUNCE_OBJECT | REDUNDANT | SPARK, 0, 0,
					     m_N_SPARK_COLORS/2 + (globalRnd.getUint32() % (m_N_SPARK_COLORS/2+1)),
					     0, 0xffffff));
				}
			}

			removeParticle( &p );
			continue;
		}

		if( part.type & SPARK ) {
			part.ttl -= part.modifier;
			if( part.ttl <= 0 ) part.ttl = 1;
			part.color = m_sparkColor[part.ttl];
		}
	}
}


Uint32 Particles::getSerializeBufferSize() const {
	return Serialize<Sint32>::sizeOf( m_nParticles ) +
		m_nParticles * Serialize<ParticleData>::sizeOf();
}


void Particles::serialize( Uint8*& bufferPointer ) const {
	START_SERIALIZED_SIZE_CHECK( bufferPointer );
	
	Serialize<Sint32>::serialize( m_nParticles, bufferPointer );

	for( int p = 0; p < m_nParticles; p++ ) {
		Serialize<ParticleData>::serialize( m_particleData[p], bufferPointer );
	}
	
	END_SERIALIZED_SIZE_CHECK( bufferPointer, Particles );
}


void Particles::deserialize( Uint8*& bufferPointer ) {
	Serialize<Sint32>::deserialize( bufferPointer, m_nParticles );

	for( int p = 0; p < m_nParticles; p++ ) {
		Serialize<ParticleData>::deserialize( bufferPointer, m_particleData[p] );
	}
}
