/* $Id: servercomm.hpp,v 1.15 2004/08/20 13:15:29 chfreund Exp $ */

#ifndef _SERVERCOMM_HPP_
#define _SERVERCOMM_HPP_

#include <vector>
#include <list>

#include "packet.hpp"
#include "comm.hpp"

class Server;
class World;
class SDL_Thread;

//! A specialized communicator class for the server
class ServerComm : public Comm {
	private:
		struct Client {
				IPaddress address;
				Event event;
				bool upToDate;
				int serverLag;
				Uint8 lastReceivedFrame;
				int clientLag;
				TCPsocket socket;
		};

	private:
		//! UDP socket which enables the client to discover the server
		UDPsocket m_baitSocket;
		//! UDP socket for the in-game communication eith the clients
		UDPsocket m_gameSocket;
		//! TCP socket for the initial synchronization of a client
		TCPsocket m_listenSocket;
		//! Socket set which will only contain the bait socket
		SDLNet_SocketSet m_baitSet;
		//! Socket set which will only contain the game socket
		SDLNet_SocketSet m_gameSet;
		//! Socket set which contains all the clients' TCP sockets
		SDLNet_SocketSet m_tcpSet;

		//! Pointer to the thread which answers incoming bait packets
		SDL_Thread* m_baitThread;
		//! Pointer to the thread which synchronizes connecting clients
		SDL_Thread* m_listenThread;

		//! Vector containg client information
		std::vector<Client*> m_client;

		//! A pointer to the server using this communicator
		Server* m_server;
		//! A pointer to the world stored on the server
		World* m_world;

		//! Status of the server
		volatile Status m_status;

		//! Buffer for the echo packets
		EchoData m_echoBuffer[N_ECHO_BUFFERS];
		
		// Players to be removed
		std::list<Uint8> m_playersToRemove;
		
	public:
		//! Constructor
		ServerComm( Server* server, World* world );
		//! Destructor
		~ServerComm();
		
		/*! \return The status of the server
		 */
		Status getStatus() const { return m_status; }
		/*! \brief Answers incoming bait packets
		 *  This function listens in an infinite loop for bait packets
		 *  from clients and answers them
		 */
		void listenForBaitPackets();
		/*! \brief Accepts incoming client connections
		 *  This function listens in an infinite loop for clients.
		 *  If a client wants to connect it also handles the initial
		 * synchronization
		 */
		void listenForClients();
		/*! \brief Gathers the events from all clients
		 */
		void collectEvents();
		/*! \brief Broadcasts the echo message to all clients
		 */
		void broadcastEcho();
		/*! \brief Handles TCP messages
		 *  Right now this concerns only chat messages
		 */
		void handleTCPMessages();
		/*! \brief Removes a client
		 *  \param c The index of the client
		 *  The connection to the client will be closed
		 */
		void removeClient( const int c );
	private:
		/*! \brief Adds a client
		 *  \param clientAddress The IP address of the new client
		 *  \param socket The socket that is connected to the client
		 */
		void addClient( const IPaddress& clientAddress, TCPsocket socket );

}; // ServerComm

/*! \brief A wrapper function for calling ServerComm::listenForClients()
 *  \param data Pointer to the ServerComm object where listenForClients()
 *    should be called
 *  \return 0 in any case
 */
int listenForClientsWrapper( void* data );
/*! \brief A wrapper function for calling ServerComm::listenForBaitPackets()
 *  \param data Pointer to the ServerComm object where listenForBaitPackets()
 *    should be called
 *  \return 0 in any case
 */
int listenForBaitPacketsWrapper( void* data );

#endif // _SERVERCOMM_HPP_
